/*
 Author: 	Raymond L. Watrous
		Department of Computer and Information Science
		University of Pennsylvania,
		Philadelphia,PA 19104
							 December, 1988.

		Department of Computer Science
		University of Toronto,
		Toronto, Ontario M5S1A4
							 September, 1990.

 Copyright (C) 1988, Trustees of the University of Pennsylvania.
 Copyright (C) 1990, Trustees of the University of Toronto.

 Permission is granted to any individual or institution to use, copy, or
 redistribute this software so long as it is not sold for profit, provided this
 copyright notice is retained. 

			   NO WARRANTY

  Because GRADSIM is licensed free of charge, absolutely NO WARRANTY is
  provided. GRADSIM is supplied "as is" without warranty of any kind, either
  expressed or implied, including, but not limited to, the implied warranties
  of merchantability and fitness for a particular purpose. The entire risk as
  to the quality and performance of the program is with you. Should the
  program prove defective, you assume the cost of all necessary servicing,
  repair or correction.

  In no event will Raymond L. Watrous, the University of Pennsylvania,
  or the University of Toronto be liable for damages, including any lost profits,
  lost monies, or other special, incidental or consequential damages arising out of
  the use or inability to use this program.

*/

/*
	apply.c

	application driver for optimization problems
	interface to function and gradient evaluations

*/



#include <stdio.h>
#include <math.h>
#include "net.h"
#include "read.h"
#include "vector.h"

#define EIGHT_BITS 255

#ifndef ENERGY_OFFSET
#define ENERGY_OFFSET 0.000
#endif

int threshold_unit;
int output_units_index[MAX_OUTPUT_UNITS];
int num_output_units;
int num_examples;
struct map_entry map[MAX_TRAIN];
int to_unit_count;

double output[MAX_DURATION + MAX_DELAY][MAX_NUM_UNITS];
#ifndef PATTERN
double increment[MAX_DURATION];
#else
long int increment[MAX_DURATION];
#endif

double slope[MAX_DURATION][MAX_NUM_UNITS];
struct link *link_set[MAX_NUM_UNITS - INPUT_UNITS];

int num_weights;

extern int read_training_descriptor();
extern char corpus[];
extern int iterations;
extern char net_name[];

#ifdef MASKED
struct link *mask_set[MAX_NUM_UNITS - INPUT_UNITS];
int mask_unit_count;
extern int mask_index[];
struct vector constant, full_gradient;
#endif

extern int to_index[],from_index[],unit_index[],delay_index[];

#ifndef PATTERN
extern double target_func();

char input[MAX_DURATION][ByteCount];

#include "io_time.c"

#else

#include "io_static.c"

#endif


int initialize(argc,argv,initial)
int argc;
char *argv[];
struct vector *initial;
{
int unit,arg_index,continuation,time;
int randomize,seed;
int type_unit_count;

#ifdef MASKED
#ifdef MULTI_UNIT_TYPE
if(argc < 5)
	{
	printf("Usage: %s [-cr] network mask unit_type experiment\n",argv[0]);
	exit(1);
	}
#else
if(argc < 4)
	{
	printf("Usage: %s [-cr] network mask experiment\n",argv[0]);
	exit(1);
	}
#endif
#else
#ifdef MULTI_UNIT_TYPE
if(argc < 4)
	{
	printf("Usage: %s [-cr] network unit_type experiment\n",argv[0]);
	exit(1);
	}
#else
if(argc < 3)
	{
	printf("Usage: %s [-cr] network experiment\n",argv[0]);
	exit(1);
	}
#endif
#endif

printf("GRADSIM 2.0 Connectionist Simulator\n");
printf("\tOptimizer %s ",argv[0]);
printf("compiled for %s\n",net_name);
for (unit=0; unit<MAX_NUM_UNITS - INPUT_UNITS; unit++)
	link_set[unit] = NULL;

continuation = 0;	arg_index = 1;
randomize = 0;

if (!strcmp(argv[arg_index],"-c"))
	{
	continuation = 1;
	arg_index++;
	}

else if (!strcmp(argv[arg_index],"-r"))
	{
	sscanf(argv[++arg_index],"%d",&seed);
	srandom(seed);
	printf("Randomization seed %d\n",seed);
	randomize = 1;
	arg_index++;
	}

printf("Optimizing network: %s using\n",argv[arg_index]);
to_unit_count = read_weights(argv[arg_index++],link_set);

#ifdef MASKED
printf("\tNetwork Mask File: %s\n",argv[arg_index]);
mask_unit_count = read_weights(argv[arg_index++],mask_set);
if (to_unit_count != mask_unit_count) exit(1);
#endif

#ifdef MULTI_UNIT_TYPE
printf("\tUnit Type Descriptor: %s\n",argv[arg_index]);
type_unit_count = read_unit_type_descriptor(argv[arg_index++],link_set,to_unit_count);
if (to_unit_count != type_unit_count) exit(1);
#endif

/*
Perturb weights if not continuation and not randomization
*/
#ifdef MASKED
if (randomize) randomize_masked_weights(link_set,mask_set);
else if (!continuation) perturb_masked_weights(link_set,mask_set);
#else
if (randomize) randomize_weights(link_set);
else if (!continuation) perturb_weights(link_set);
#endif

printf("\tExperiment: %s\n",argv[arg_index]);
read_training_descriptor(argv[arg_index]);

for (time=0; time< MAX_DELAY; time++)
	for (unit=0; unit<MAX_NUM_UNITS; unit++)
		output[time][unit] = ENERGY_OFFSET;

for (time=0; time<MAX_DURATION + MAX_DELAY; time++)
	output[time][threshold_unit] = 1.000;

#ifdef MASKED
num_weights = map_net_vect(link_set,to_unit_count,&constant);

return(map_mask_vect(link_set,mask_set,to_unit_count,initial));
#else
num_weights = map_net_vect(link_set,to_unit_count,initial);
return(num_weights);
#endif
}



double objective(filename,weight)
char filename[];
struct vector *weight;
/*
Evaluates network over corpus based on target
*/
{
int example;
struct map_entry *descriptor;
register int length,out,unit,time,otime;
register double total,extent,target,error,average;
register double (*od)[MAX_NUM_UNITS];
double epsilon;
#ifndef PATTERN
double increment;
#endif

total = 0;
for (example=0; example<num_examples; example++)
	{
	descriptor = &map[example];
	average = 0;
#ifndef PATTERN
	length = read_in_data(descriptor,output[MAX_DELAY]);
#else
	length = read_in_data(descriptor,output[MAX_DELAY],increment);
#endif
	forward_pass(length,weight,output[MAX_DELAY]);

	for (time=0,otime=MAX_DELAY; time<length; time++,otime++)
		{
#ifndef PATTERN
		extent = ((double) time) / length;
		increment = target_func(extent,descriptor->xi,
			descriptor->mag);
#ifdef INTERMITTENT
		if (increment >= 0.00) {
#endif
#endif
		for (out=0; out<num_output_units; out++)
			{
			unit = output_units_index[out];
#ifndef PATTERN
#ifdef ASYMMETRIC
			if (descriptor->output_unit_sign[out] == 1.00)
				target = 0.05 + 2 * increment;
			else target = 0.05;
#else
			if (descriptor->output_unit_sign[out] == 1.00)
				target = 0.500 + increment;
			else target = 0.500 - increment;
#endif
#else
#ifdef EXPLICIT
			target = descriptor->output_unit_sign[out];
#else
#ifdef ASCII_CLASS_DATA
			if (increment[time] == out) target = 0.900;
			else target = 0.100;
#else
			if (descriptor->output_unit_sign[out] == 1.00) target = 0.900;
			else target = 0.100;
#endif
#endif
#endif
			error = (target - output[otime][unit]);
			average += error * error;
			}
#ifndef PATTERN
#ifdef INTERMITTENT
}
#endif
#endif
		}

#ifdef RADIAL
if (descriptor->xi != 0.000) average /= descriptor->xi;
#endif

#ifdef LMS
if (length != 0) total += average/length;
#else
total += average;
#endif
	}
#ifdef LMS
return(total/(2 * num_output_units * num_examples)); 
#else
return(total/2);
#endif
}




/* CLEAR VECT */

int clear_vect(vect)
struct vector *vect;
{
register double *dp;
register int index;

dp = vect->val;
for (index=0; index<num_weights; index++)
	*dp++ = 0.00;
}


/* NORM VECT */

int norm_vect(vect,count)
struct vector *vect;
int count;
{
register double scale;
register double *dp;
register int index;

scale = count;
dp = vect->val;
for (index=0; index<num_weights; index++)
	*dp++ /= scale;
}



/* FIND GRADIENT */

int find_gradient(filename,weight,grad_vect)
char filename[];
struct vector *weight,*grad_vect;
/*
Computes gradient vector from corpus, for given network
*/
{
register int time,length,example;
register double extent;
register struct map_entry *descriptor;

clear_vect(grad_vect);
for (example=0; example<num_examples; example++)
	{
	descriptor = &map[example];
#ifndef PATTERN
	length = read_in_data(descriptor,output[MAX_DELAY]);
#else
	length = read_in_data(descriptor,output[MAX_DELAY],increment);
#endif
	forward_pass(length,weight,output[MAX_DELAY]);
#ifndef PATTERN
	for (time=0; time<length; time++)
		{
		extent = ((double) time) / length;
		increment[time] = target_func(extent,
			descriptor->xi, descriptor->mag);
		}
#endif
	cg3 (output[MAX_DELAY],grad_vect,weight,
		descriptor,length,increment);
	}
#ifdef LMS
norm_vect(grad_vect,num_examples);
#endif
}

