/*
 Author: 	Raymond L. Watrous
		Department of Computer and Information Science
		University of Pennsylvania,
		Philadelphia,PA 19104
							 December, 1988.

		Department of Computer Science
		University of Toronto,
		Toronto, Ontario M5S1A4
							 September, 1990.

 Copyright (C) 1988, Trustees of the University of Pennsylvania.
 Copyright (C) 1990, Trustees of the University of Toronto.

 Permission is granted to any individual or institution to use, copy, or
 redistribute this software so long as it is not sold for profit, provided this
 copyright notice is retained. 

			   NO WARRANTY

  Because GRADSIM is licensed free of charge, absolutely NO WARRANTY is
  provided. GRADSIM is supplied "as is" without warranty of any kind, either
  expressed or implied, including, but not limited to, the implied warranties
  of merchantability and fitness for a particular purpose. The entire risk as
  to the quality and performance of the program is with you. Should the
  program prove defective, you assume the cost of all necessary servicing,
  repair or correction.

  In no event will Raymond L. Watrous, the University of Pennsylvania,
  or the University of Toronto be liable for damages, including any lost profits,
  lost monies, or other special, incidental or consequential damages arising out of
  the use or inability to use this program.

*/

/*
	bp.c

		Back-Propagation algorithm
	for connectionist networks. 
	Basically, compute gradient at point x, move fixed amount
	in negative gradient direction and repeat.

	11/28/87 - modified for iteration count

	11/30/87 - modified based on vector version

	1/20/88 - modified for compute_gradient changes
*/

#include <stdio.h>
#include <string.h>
#include "vector.h"

extern int initialize();
extern int space;
extern double objective_function();
extern int compute_gradient();

#define EPSILON  0.00000001
#define ETA 0.15
#define MU 0.2

#define MAX_ITS 50000

#define NUM_GOALS 20
#define ACCEPTABLE 0.0000000001
#define BASE_GOAL 0.000002

#ifdef TERMIN
double goal[NUM_GOALS];
int goal_index;
#endif

char corpus[100];
int iterations;
int function_evaluations;
int gradient_evaluations;

int space;			/* global space size */



int update_vect(a,b,c,s,m)
struct vector *a, *b, *c;
double s,m;
{
int index;
double increment;

for (index=0; index<space; index++)
	{
	increment = m * c->val[index] - s * b->val[index];
	a->val[index] += increment;
	c->val[index] = increment;
	}
iterations++;
} 




main(argc,argv)
int argc;
char *argv[];
{
struct vector x0, grad_vect, change_hist;
double alpha,objective_value,min,mu,best;
char net_name[100];
int done,index;

space = initialize(argc,argv,&x0);

for (index=0; index<space; index++)
	change_hist.val[index] = 0.00;

#ifdef TERMIN
for (goal_index=0; goal_index<NUM_GOALS; goal_index++)
	goal[goal_index] = BASE_GOAL * (1.00 - ((double) goal_index)/NUM_GOALS);
goal_index = 0;
#endif

iterations = 0;
done = 0;
min = ACCEPTABLE;
alpha = ETA;
mu = MU;
best = 10000000000000;
printf("Alpha %.10f\tMu %.10f\n",alpha,mu);
while (!done)
	{
	objective_value = objective_function(&x0);
	printf("%d\t%d\t%d\tObjective value\t%f\n",iterations,
		function_evaluations,
		gradient_evaluations,
		objective_value); 
	fflush(stdout);
	if (objective_value < best)
		{
#ifdef TERMIN
		while ((goal_index < NUM_GOALS) && (objective_value <goal[goal_index]))
			{
			printf("Met Goal %f:\t%d\t%d\t%d\n",goal[goal_index],iterations,
				function_evaluations,gradient_evaluations);
			sprintf(net_name,"net.t%f.fin",goal[goal_index]);
			finalize(net_name,&x0);
			goal_index++;
			}
#else
		finalize("net.tmp.tmp",&x0);
#endif
		best = objective_value;
		}
	if ((objective_value < min) || (iterations>MAX_ITS)) done = 1;
	else
		{
		compute_gradient(&x0,&grad_vect);
		update_vect(&x0,&grad_vect,&change_hist,alpha,mu);
		}
	}
printf("Best Value = %f\n",best);
if (objective_value < min)
printf("Iterations to solve = %d\n",iterations);
else printf("Iterations to failure = %d\n",iterations);
finalize("net.x.fin",&x0);
}


