/*
 Author: 	Raymond L. Watrous
		Department of Computer and Information Science
		University of Pennsylvania,
		Philadelphia,PA 19104
							 December, 1988.

		Department of Computer Science
		University of Toronto,
		Toronto, Ontario M5S1A4
							 September, 1990.

 Copyright (C) 1988, Trustees of the University of Pennsylvania.
 Copyright (C) 1990, Trustees of the University of Toronto.

 Permission is granted to any individual or institution to use, copy, or
 redistribute this software so long as it is not sold for profit, provided this
 copyright notice is retained. 

			   NO WARRANTY

  Because GRADSIM is licensed free of charge, absolutely NO WARRANTY is
  provided. GRADSIM is supplied "as is" without warranty of any kind, either
  expressed or implied, including, but not limited to, the implied warranties
  of merchantability and fitness for a particular purpose. The entire risk as
  to the quality and performance of the program is with you. Should the
  program prove defective, you assume the cost of all necessary servicing,
  repair or correction.

  In no event will Raymond L. Watrous, the University of Pennsylvania,
  or the University of Toronto be liable for damages, including any lost profits,
  lost monies, or other special, incidental or consequential damages arising out of
  the use or inability to use this program.

*/

/*

	conjugate.c

		Implementation of Polak-Ribiere conjugate gradient
	algorithm


*/


#include <stdio.h>
#include <math.h>
#include <string.h>
#include "vector.h"

extern int initialize();
extern int finalize();
extern double objective_function();
extern int compute_gradient();
extern double find_min_along();
extern double dot_product();

extern int space;

#define EPSILON 0.000000000001

#define ACCEPTABLE 0.00000001

#define NUM_GOALS 20

#define BASE_GOAL 0.010

#ifdef TERMIN
double goal[NUM_GOALS];
int gndx;
#endif


char corpus[100];
int iterations;
int function_evaluations;
int gradient_evaluations;


struct vector delta;


int neg_vect(a,b)
struct vector *a, *b;
{
int index;

for (index=0; index<space; index++)
	a->val[index] = -(b->val[index]);
} 



int vector_sum(a,b,c)
/* a gets vector b + c */
struct vector *a, *b, *c;
{
int index;

for (index=0; index<space; index++)
	a->val[index] = b->val[index] + c->val[index];
} 



int vector_difference(a,b,c)
struct vector *a, *b, *c;
{
int index;

for (index=0; index<space; index++)
	a->val[index] = b->val[index] - c->val[index];
} 



int vector_scale(a,b,s)
struct vector *a, *b;
double s;
{
int index;

for (index=0; index<space; index++)
	a->val[index] = b->val[index] * s;
} 





int update_polak(x,s,g,alpha,gnew)
struct vector *x, *s, *g, *gnew;
double alpha;
{
double beta, delgam, gmag ;

/*
Side effects - updates x and s
*/
vector_scale(&delta,s,alpha);
vector_sum(x,x,&delta);
#ifndef FLETCHER_REEVES
vector_difference(&delta,gnew,g);
delgam = dot_product(&delta,gnew);
#else
delgam = dot_product(gnew,gnew);
#endif

gmag = dot_product(g,g);
beta = delgam/gmag;

neg_vect(&delta,gnew);
vector_scale(s,s,beta);
vector_sum(s,&delta,s);
iterations++;
}




main(argc,argv)
int argc;
char *argv[];
{
static struct vector x0;
struct vector grad1,grad2,search_vect,v1;
struct vector *gradient,*new_grad, *temp;
double alpha,objective_value,min;
int done,strikes,i;
char net_name[100];


space = initialize(argc,argv,&x0);

finalize("net.testio.fin",&x0);

iterations = 0;
done = 0;
min = ACCEPTABLE;

objective_value = objective_function(&x0);
printf("%d\tObjective value\t%.10f\n",iterations,objective_value); 
gradient = &grad1;
new_grad = &grad2;

#ifdef TERMIN
for (gndx=0; gndx<NUM_GOALS; gndx++)
	goal[gndx] = BASE_GOAL * (1.00 - ((double) gndx)/NUM_GOALS);

if (objective_value < goal[NUM_GOALS-1]) done = 1;
else
	{
	compute_gradient(&x0,gradient);
	neg_vect(&search_vect,gradient);
	}

gndx = 0;
#else
if (objective_value < min) done = 1;
else 
	{
	compute_gradient(&x0,gradient);
	neg_vect(&search_vect,gradient);
	}
#endif

while (!done)
	{
#ifdef TERMIN
	while ((gndx < NUM_GOALS) && (objective_value <goal[gndx]))
		{
		printf("Goal Index %d\t",gndx);
		printf("Met Goal %f:\t%d\t%d\t%d\n",goal[gndx],iterations,
			function_evaluations,gradient_evaluations);
		sprintf(net_name,"net.t%f.fin",goal[gndx]);
		finalize(net_name,&x0);
		gndx++;
		}
	if (gndx == NUM_GOALS) done = 1;
#else
	if (objective_value < min) done = 1;
#endif
	else
		{
		alpha = find_min_along(&objective_value,&x0,
			gradient,&search_vect,objective_function,
			compute_gradient,new_grad);
#ifndef SILENT
		printf("%d\t%d\t%d\tObjective value\t%.10f\t",iterations,
			function_evaluations,
			gradient_evaluations,
			objective_value); 
		printf("Alpha %.10f\n",alpha);
		fflush(stdout);
#endif
		if (alpha == 0.000) done = 1;
#ifdef RESTART
		if (iterations % space == 0)
			{
			vector_scale(&delta,&search_vect,alpha);
			vector_sum(&x0,&x0,&delta);
			neg_vect(&search_vect,new_grad);
			iterations++;
			}
		else
#endif
			{
			update_polak(&x0,&search_vect,gradient,alpha,new_grad);
			}

/*
SWITCH GRADIENTS
*/
		temp = gradient;
		gradient = new_grad;
		new_grad = temp;

		finalize("net.tmp.tmp",&x0);
		}
	}
printf("%d\t%d\t%d\tObjective value\t%.10f\n",iterations,
	function_evaluations,
	gradient_evaluations,
	objective_value); 
if (objective_value < min)
printf("Iterations to solve = %d\n",iterations);
else printf("Iterations to failure = %d\n",iterations);
finalize("net.x.fin",&x0);
}

