/*
 Author: 	Raymond L. Watrous
		Department of Computer and Information Science
		University of Pennsylvania,
		Philadelphia,PA 19104
							 December, 1988.

		Department of Computer Science
		University of Toronto,
		Toronto, Ontario M5S1A4
							 September, 1990.

 Copyright (C) 1988, Trustees of the University of Pennsylvania.
 Copyright (C) 1990, Trustees of the University of Toronto.

 Permission is granted to any individual or institution to use, copy, or
 redistribute this software so long as it is not sold for profit, provided this
 copyright notice is retained. 

			   NO WARRANTY

  Because GRADSIM is licensed free of charge, absolutely NO WARRANTY is
  provided. GRADSIM is supplied "as is" without warranty of any kind, either
  expressed or implied, including, but not limited to, the implied warranties
  of merchantability and fitness for a particular purpose. The entire risk as
  to the quality and performance of the program is with you. Should the
  program prove defective, you assume the cost of all necessary servicing,
  repair or correction.

  In no event will Raymond L. Watrous, the University of Pennsylvania,
  or the University of Toronto be liable for damages, including any lost profits,
  lost monies, or other special, incidental or consequential damages arising out of
  the use or inability to use this program.

*/

/*
	grad.c

		Gradient descent, i.e. steepest descent
	for connectionist networks using line-search.

	Basically, compute gradient at point x, call line
	search and update
	terminate upon sufficient condition

	1/20/88 - modified for compute_gradient changes

*/

#include <stdio.h>
#include <string.h>
#include "vector.h"

extern int initialize();
extern int space;
extern double objective_function();
extern int compute_gradient();
extern double find_min_along();

#define EPSILON  0.00000001

#define ACCEPTABLE 0.00000001

#ifdef TERMIN
#define NUM_GOALS 20

#define BASE_GOAL 0.010

double goal[NUM_GOALS];
int gndx;
#endif

char corpus[100];

int iterations;
int function_evaluations;
int gradient_evaluations;


int negate_vect(a,b)
struct vector *a,*b;
{
int index;

for (index=0; index<space; index++)
	a->val[index] = -(b->val[index]);
} 


int update_vect(a,b,s)
struct vector *a, *b;
double s;
{
int index;

for (index=0; index<space; index++)
	a->val[index] = a->val[index] + s * b->val[index];
iterations++;
} 






main(argc,argv)
int argc;
char *argv[];
{
struct vector x0, grad1,grad2,search_vect;
struct vector *grad_vect, *new_grad, *temp;
double alpha,objective_value,min;
int done;
char net_name[100];

space = initialize(argc,argv,&x0);

grad_vect = &grad1;
new_grad = &grad2;

iterations = 0;
done = 0;
min = ACCEPTABLE;
objective_value = objective_function(&x0);
printf("%d\tObjective value\t%f\n",iterations,objective_value); 

#ifdef TERMIN
for (gndx=0; gndx<NUM_GOALS; gndx++)
	goal[gndx] = BASE_GOAL * (1.00 - ((double) gndx)/NUM_GOALS);

if (objective_value < goal[NUM_GOALS-1]) done = 1;
else compute_gradient(&x0,grad_vect);
gndx = 0;
#else
if (objective_value < min) done = 1;
else compute_gradient(&x0,grad_vect);
#endif
while (!done)
	{
	negate_vect(&search_vect,grad_vect);
	alpha = find_min_along(&objective_value,&x0,
		grad_vect,&search_vect,objective_function,
		compute_gradient,new_grad);
	printf("%d\t%d\t%d\tObjective value\t%.10f\t",iterations,
		function_evaluations,
		gradient_evaluations,
		objective_value); 
	printf("Alpha %.10f\n",alpha);
	fflush(stdout);
#ifdef TERMIN
	while ((gndx < NUM_GOALS) && (objective_value <goal[gndx]))
		{
		printf("Goal Index %d\t",gndx);
		printf("Met Goal %f:\t%d\t%d\t%d\n",goal[gndx],iterations,
			function_evaluations,gradient_evaluations);
		sprintf(net_name,"net.t%f.fin",goal[gndx]);
		finalize(net_name,&x0);
		gndx++;
		}
	if (gndx == NUM_GOALS) done = 1;
#else
	if ((alpha == 0.000) || (objective_value < min)) done = 1;
#endif
	else
		{
		update_vect(&x0,&search_vect,alpha);
		finalize("net.tmp.tmp",&x0);
		temp = grad_vect;
		grad_vect = new_grad;
		new_grad = temp;
		}
	}
if (objective_value < min)
printf("Iterations to solve = %d\n",iterations);
else printf("Iterations to failure = %d\n",iterations);
finalize("net.x.fin",&x0);
}


