/*
 Author: 	Raymond L. Watrous
		Department of Computer and Information Science
		University of Pennsylvania,
		Philadelphia,PA 19104
							 December, 1988.

		Department of Computer Science
		University of Toronto,
		Toronto, Ontario M5S1A4
							 September, 1990.

 Copyright (C) 1988, Trustees of the University of Pennsylvania.
 Copyright (C) 1990, Trustees of the University of Toronto.

 Permission is granted to any individual or institution to use, copy, or
 redistribute this software so long as it is not sold for profit, provided this
 copyright notice is retained. 

			   NO WARRANTY

  Because GRADSIM is licensed free of charge, absolutely NO WARRANTY is
  provided. GRADSIM is supplied "as is" without warranty of any kind, either
  expressed or implied, including, but not limited to, the implied warranties
  of merchantability and fitness for a particular purpose. The entire risk as
  to the quality and performance of the program is with you. Should the
  program prove defective, you assume the cost of all necessary servicing,
  repair or correction.

  In no event will Raymond L. Watrous, the University of Pennsylvania,
  or the University of Toronto be liable for damages, including any lost profits,
  lost monies, or other special, incidental or consequential damages arising out of
  the use or inability to use this program.

*/

/*

	map.c

	maps vectors to network structures and reverse

	5/18/87 - modified for direct vector processing

	11/3/87 - included delay_index

	12/15/87 - modified to check MAX_DELAY

	1/20/88 - modified to add argument to compute_gradient

*/

#include <stdio.h>
#include "net.h"
#include "vector.h"


extern int function_evaluations;
extern int gradient_evaluations;
extern char corpus[];
extern struct link *link_set[];
extern int to_unit_count;
extern int space;
extern int find_gradient();
extern double objective();

int to_index[MAX_VECTOR];
int from_index[MAX_VECTOR];
int delay_index[MAX_VECTOR];
int unit_index[MAX_NUM_UNITS];
#ifdef SIGPI
int f2_index[MAX_VECTOR];
int d2_index[MAX_VECTOR];
int link_type[MAX_VECTOR];
#endif
#ifdef MASKED
int mask_index[MAX_VECTOR];
extern struct vector constant, full_gradient;
#endif

#ifdef MULTI_UNIT_TYPE
int unit_type[MAX_NUM_UNITS];
#endif



int map_net_vect(link_set,count,vp)
struct link *link_set[];
int count;
struct vector *vp;
{
int unit_ndx,vector_index;
struct link *lp;

vector_index = 0;
for (unit_ndx=0; unit_ndx <count; unit_ndx++)
	{
	lp = link_set[unit_ndx];
	unit_index[unit_ndx] = lp->to_unit;
	while (lp != NULL)
		{
		to_index[vector_index] = lp->to_unit;
		from_index[vector_index] = lp->from_unit;
		delay_index[vector_index] = lp->delay;
#ifdef SIGPI
		if (lp->type == SIGMA_PI)
			{
			f2_index[vector_index] = lp->from2_unit;
			d2_index[vector_index] = lp->delay2;
			link_type[vector_index] = SIGMA_PI;
			}		  
		else link_type[vector_index] = SIMPLE;
#endif
		vp->val[vector_index++] = lp->weight;
		lp = lp->next;
		}
	}
if (vector_index > MAX_VECTOR)
	{
	fprintf(stderr,"Link Count %d exceeds MAX_VECTOR - leaving\n",
		vector_index);
	exit(1);
	}
return(vector_index);
}




int map_vect_net(vp,ls,count)
struct vector *vp;
struct link *ls[];
int count;
{
int unit_ndx,vector_index;
struct link *lp;

vector_index = 0;
for (unit_ndx=0; unit_ndx <count; unit_ndx++)
	{
	lp = ls[unit_ndx];
	while (lp != NULL)
		{
		lp->weight = vp->val[vector_index++];
		lp = lp->next;
		}
	}
return(vector_index);
}


#ifdef MASKED

int map_mask_vect(link_set,mask_set,count,vp)
struct link *link_set[];
struct link *mask_set[];
int count;
struct vector *vp;
/*
maps mask structure onto vector structure, uses 1.000 in weight field as mask allowing
variable weight. Assumes that link and mask structure have been correctly read in and
match. Also generates mask index structure to relate variable to fixed weight structures.
*/{
int unit_ndx,vector_index;
int relative;
struct link *lp;
struct link *mp;

vector_index = 0;
relative = 0;
for (unit_ndx=0; unit_ndx <count; unit_ndx++)
	{
	lp = link_set[unit_ndx];
	mp = mask_set[unit_ndx];
	while (lp != NULL)
		{
		if (mp->weight == 1.000) 
			{
			vp->val[vector_index] = lp->weight;
			mask_index[vector_index++] = relative;
			}
		relative++;
		lp = lp->next;
		mp = mp->next;
		}
	}
return(vector_index);
}


map_var_fixed(mv,cv,n)
struct vector *mv, *cv;
int n;
/*
Maps n masked vector (mv) variables onto constant vector (cv) using mask index
*/

{
int i;

for (i=0; i<n; i++)
	cv->val[mask_index[i]] = mv->val[i];
}


map_fixed_var(cv,mv,n)
struct vector *cv, *mv;
int n;
/*
Maps n masked vector (mv) variables onto constant vector (cv) using mask index
*/

{
int i;

for (i=0; i<n; i++)
	mv->val[i] = cv->val[mask_index[i]];
}



#endif


double objective_function(vect)
struct vector *vect;
{
/*
call network evaluation function based on corpus
*/
function_evaluations++;

#ifdef MASKED
map_var_fixed(vect,&constant,space);
return(objective(corpus,&constant));
#else
return(objective(corpus,vect));
#endif
}



int compute_gradient(vect,grad)
struct vector *vect,*grad;
{
/*
call gradient evaluation function using weight vector and grad as result
*/
gradient_evaluations++;
#ifdef MASKED
find_gradient(corpus,&constant,&full_gradient);
map_fixed_var(&full_gradient,grad,space);
#else
find_gradient(corpus,vect,grad);
#endif
}




int finalize(network_name,final)
char *network_name;
struct vector *final;
{
#ifdef MASKED
map_var_fixed(final,&constant,space);
map_vect_net(&constant,link_set,to_unit_count);
#else
map_vect_net(final,link_set,to_unit_count);
#endif
store_weights(network_name,link_set);
}



