/*
 Author: 	Raymond L. Watrous
		Department of Computer and Information Science
		University of Pennsylvania,
		Philadelphia,PA 19104
							 December, 1988.

		Department of Computer Science
		University of Toronto,
		Toronto, Ontario M5S1A4
							 September, 1990.

 Copyright (C) 1988, Trustees of the University of Pennsylvania.
 Copyright (C) 1990, Trustees of the University of Toronto.

 Permission is granted to any individual or institution to use, copy, or
 redistribute this software so long as it is not sold for profit, provided this
 copyright notice is retained. 

			   NO WARRANTY

  Because GRADSIM is licensed free of charge, absolutely NO WARRANTY is
  provided. GRADSIM is supplied "as is" without warranty of any kind, either
  expressed or implied, including, but not limited to, the implied warranties
  of merchantability and fitness for a particular purpose. The entire risk as
  to the quality and performance of the program is with you. Should the
  program prove defective, you assume the cost of all necessary servicing,
  repair or correction.

  In no event will Raymond L. Watrous, the University of Pennsylvania,
  or the University of Toronto be liable for damages, including any lost profits,
  lost monies, or other special, incidental or consequential damages arising out of
  the use or inability to use this program.

*/

/*
	perturb.c

	perturbs initial weight configuration
	by adding random epsilon value less than
	0.1% of initial value

	modified 10/30/87
		corrected BigNum to (2^31 -1)
		use to_unit_count instead of highest_unit_index
*/

#include <stdio.h>
#include "net.h"

extern int to_unit_count;

extern struct link *link_set[];

#define BigNum 2.147483647e9		/* (2^31-1) for random */


float scaled_random(range)
double range;
/* Computes random value between 0 and range */
{
double value;

value = (random() * range)/BigNum;
return(value);
}


float perturb(weight)
float weight;
/* Return a weight perturbed by a small random amount <0.1% of initial value */

{
return(weight + scaled_random(weight/1000));
}


int perturb_weights(ls)
struct link *ls[];
{
int unit;
struct link *lp;

for (unit=0; unit<to_unit_count; unit++)
	{
	lp = ls[unit];
	while (lp != NULL)
		{
		lp->weight = perturb(lp->weight);
		lp = lp->next;
		}
	}
}


/*
Randomize Weights between -1 and +1
*/
int randomize_weights(ls)
struct link *ls[];
{
int unit;
struct link *lp;

for (unit=0; unit<to_unit_count; unit++)
	{
	lp = ls[unit];
	while (lp != NULL)
		{
		lp->weight = 1.000 - scaled_random(2.00);
		lp = lp->next;
		}
	}
}


#ifdef MASKED

/*
Randomize Weights between -1 and +1
*/
int randomize_masked_weights(ls,ms)
struct link *ls[], *ms[];
{
int unit;
struct link *lp, *mp;

for (unit=0; unit<to_unit_count; unit++)
	{
	lp = ls[unit];
	mp = ms[unit];
	while (lp != NULL)
		{
		if (mp->weight == 1.000) lp->weight = 1.000 - scaled_random(2.00);
		lp = lp->next;
		}
	}
}



int perturb_masked_weights(ls,ms)
struct link *ls[], *ms[];
{
int unit;
struct link *lp, *mp;

for (unit=0; unit<to_unit_count; unit++)
	{
	lp = ls[unit];
	mp = ms[unit];
	while (lp != NULL)
		{
		if (mp->weight == 1.000) perturb(lp->weight);
		lp = lp->next;
		}
	}
}


#endif

