/*
 Author: 	Raymond L. Watrous
		Department of Computer and Information Science
		University of Pennsylvania,
		Philadelphia,PA 19104
							 December, 1988.

		Department of Computer Science
		University of Toronto,
		Toronto, Ontario M5S1A4
							 September, 1990.

 Copyright (C) 1988, Trustees of the University of Pennsylvania.
 Copyright (C) 1990, Trustees of the University of Toronto.

 Permission is granted to any individual or institution to use, copy, or
 redistribute this software so long as it is not sold for profit, provided this
 copyright notice is retained. 

			   NO WARRANTY

  Because GRADSIM is licensed free of charge, absolutely NO WARRANTY is
  provided. GRADSIM is supplied "as is" without warranty of any kind, either
  expressed or implied, including, but not limited to, the implied warranties
  of merchantability and fitness for a particular purpose. The entire risk as
  to the quality and performance of the program is with you. Should the
  program prove defective, you assume the cost of all necessary servicing,
  repair or correction.

  In no event will Raymond L. Watrous, the University of Pennsylvania,
  or the University of Toronto be liable for damages, including any lost profits,
  lost monies, or other special, incidental or consequential damages arising out of
  the use or inability to use this program.

*/

/*
        read.c
                reads in training descriptor

	4/5/90 - add explicit target

*/


#include <stdio.h>
#include "read.h"

extern int output_units_index[];

extern int num_output_units;

extern int num_examples;

extern struct map_entry map[];


int read_training_descriptor(name)
char *name;
{
int entry,i;
int x;
struct map_entry *mp;
char temp[MAX_NAME_LENGTH];
FILE *fp;

if ((fp = fopen(name,"r")) == NULL)
	{
	printf("CANNOT OPEN EXPERIMENT DESCRIPTOR");
	exit(1);
	}
fscanf(fp," %d",&num_output_units);

if (num_output_units > MAX_OUTPUT_UNITS)
	{
	fprintf(stderr,"***MORE OUTPUT UNITS THAN ALLOWED***\n");
	exit(1);
	}

for (i=0; i<num_output_units; i++)
	entry = fscanf(fp," %d",&output_units_index[i]);
entry = 0;
while (((x = fscanf(fp," %s",temp)) != EOF) && (entry<MAX_TRAIN))
	{
	mp = &map[entry++];
	strcpy(mp->filename,temp);
	for (i=0; i<num_output_units; i++)
		{
#ifdef EXPLICIT
		fscanf(fp,"%f",&mp->output_unit_sign[i]);
#else
		fscanf(fp," %s",temp);
		if (!strcmp("+",temp))
			mp->output_unit_sign[i] = 1.0000;
		else mp->output_unit_sign[i] = -1.0000;
#endif
		}
	fscanf(fp,"%f %f",&mp->xi,&mp->mag);
#ifdef RADIAL
	if (mp->xi == 0.000)
		{
		printf("ZERO VARIANCE FOR RADIAL BASIS FUNCTION");
		exit(1);
		}
#endif
#ifdef CONTEXT
	fscanf(fp,"%d",&mp->context);
#endif
	}
fclose(fp);
num_examples = entry;
return(entry);
}




int write_training_descriptor(name)
char *name;
{
int entry,i;
int x;
struct map_entry *mp;
FILE *fp;

if ((fp = fopen(name,"w")) == NULL)
	{
	printf("CANNOT OPEN TRAINING DESCRIPTOR");
	exit(1);
	}
fprintf(fp,"%d",num_output_units);
for (i=0; i<num_output_units; i++)
	fprintf(fp,"\t%d",output_units_index[i]);
fprintf(fp,"\n");
for (entry=0; entry<num_examples; entry++)
	{
	mp = &map[entry];
	fprintf(fp,"%s",mp->filename);
	for (i=0; i<num_output_units; i++)
		{
#ifdef EXPLICIT
		fprintf(fp,"\t%f",mp->output_unit_sign[i]);
#else
		if (mp->output_unit_sign[i] == 1.00)
			fprintf(fp,"\t+");
		else fprintf(fp,"\t-");
#endif
		}
	fprintf(fp,"\t%f\t%f\t",mp->xi,mp->mag);
#ifdef CONTEXT
        fprintf(fp,"%d\n",&mp->context);
#endif
	}
fclose(fp);
}



