/*
 Author: 	Raymond L. Watrous
		Department of Computer and Information Science
		University of Pennsylvania,
		Philadelphia,PA 19104
							 December, 1988.

		Department of Computer Science
		University of Toronto,
		Toronto, Ontario M5S1A4
							 September, 1990.

 Copyright (C) 1988, Trustees of the University of Pennsylvania.
 Copyright (C) 1990, Trustees of the University of Toronto.

 Permission is granted to any individual or institution to use, copy, or
 redistribute this software so long as it is not sold for profit, provided this
 copyright notice is retained. 

			   NO WARRANTY

  Because GRADSIM is licensed free of charge, absolutely NO WARRANTY is
  provided. GRADSIM is supplied "as is" without warranty of any kind, either
  expressed or implied, including, but not limited to, the implied warranties
  of merchantability and fitness for a particular purpose. The entire risk as
  to the quality and performance of the program is with you. Should the
  program prove defective, you assume the cost of all necessary servicing,
  repair or correction.

  In no event will Raymond L. Watrous, the University of Pennsylvania,
  or the University of Toronto be liable for damages, including any lost profits,
  lost monies, or other special, incidental or consequential damages arising out of
  the use or inability to use this program.

*/

/*

	rosen.c

		simple test program for checking optimization
	algorithms using Rosenbrock's function


*/

#include "vector.h"

extern int function_evaluations;
extern int gradient_evaluations;



#define BigNum 2.147483647e9		/* (2^31-1) for random */


float scaled_random(range)
double range;
/* Computes random value in +- range */
{
double value;

value = 1.000 - 2 * (random() * range)/BigNum;
return(value);
}


int initialize(argc,argv,initial)
int argc;
char *argv[];
struct vector *initial;
{
/* MUST REMAIN FLOAT!!*/
float val;
int seed,randomize;

if(argc != 3)
	{
	printf("Usage: %s [-r seed] [x1  x2]\n",argv[0]);
	exit(1);
	}

randomize = 0;

if (!strcmp(argv[1],"-r"))
	{
	sscanf(argv[2],"%d",&seed);
	srandom(seed);
	printf("Randomization seed %d\n",seed);
	randomize = 1;
	}
if (randomize)
	{
	initial->val[0] = scaled_random(10.00);
	initial->val[1] = scaled_random(10.00);
	}
else
	{
/*
sscanf(argv[1],"%f",&initial->val[0]);
sscanf(argv[2],"%f",&initial->val[1]);
*/
	sscanf(argv[1],"%f",&val);
	initial->val[0] = val;
	sscanf(argv[2],"%f",&val);
	initial->val[1] = val;
	}
return(2);
}



double objective_function(x)
struct vector *x;
{
double total,val;
int i;

function_evaluations++;
total = 0;
val = (1- x->val[0]);
total += val * val;
val = (x->val[1] - x->val[0] * x->val[0]);
total += (100 * val * val);
return(total);
}



int compute_gradient(x,g)
struct vector *x,*g;
{
double val;

gradient_evaluations++;
val = (x->val[1] - x->val[0] * x->val[0]);
g->val[0] = -400 * val * x->val[0] - 2 * (1 - x->val[0]);
g->val[1] = 200 * val;
}



int finalize(network_name,final)
char *network_name;
struct vector *final;
{
#ifdef UPDATE
printf("x1 = %f\nx2 = %f\n",final->val[0],final->val[1]);
#endif
}


