/*
 Author: 	Raymond L. Watrous
		Department of Computer and Information Science
		University of Pennsylvania,
		Philadelphia,PA 19104
							 December, 1988.

		Department of Computer Science
		University of Toronto,
		Toronto, Ontario M5S1A4
							 September, 1990.

 Copyright (C) 1988, Trustees of the University of Pennsylvania.
 Copyright (C) 1990, Trustees of the University of Toronto.

 Permission is granted to any individual or institution to use, copy, or
 redistribute this software so long as it is not sold for profit, provided this
 copyright notice is retained. 

			   NO WARRANTY

  Because GRADSIM is licensed free of charge, absolutely NO WARRANTY is
  provided. GRADSIM is supplied "as is" without warranty of any kind, either
  expressed or implied, including, but not limited to, the implied warranties
  of merchantability and fitness for a particular purpose. The entire risk as
  to the quality and performance of the program is with you. Should the
  program prove defective, you assume the cost of all necessary servicing,
  repair or correction.

  In no event will Raymond L. Watrous, the University of Pennsylvania,
  or the University of Toronto be liable for damages, including any lost profits,
  lost monies, or other special, incidental or consequential damages arising out of
  the use or inability to use this program.

*/

/*

	score.c

		for scoring nets

*/


#include <stdio.h>
#include <math.h>
#include <string.h>
#include "vector.h"
#include "read.h"
#include "net.h"

int space;
int function_evaluations, gradient_evaluations;

float perturb_weights(){}

int randomize_weights(){}

char corpus[10];

extern int num_examples;
extern int num_output_units;
extern int output_units_index[];
extern struct map_entry map[];
extern double output[MAX_DURATION + MAX_DELAY][MAX_NUM_UNITS];


#ifndef PATTERN
#define PLUS 0
#define MINUS 1
extern double target_func();
extern double increment[];
#else
extern long int increment[];
#endif


score_network(w)
struct vector *w;
{
int j;
struct map_entry *mp;
int x,length;
int out,unit,index;
double dat,max;
#ifndef PATTERN
int i,test_unit;
int hyp, argmin;
double score[MAX_OUTPUT_UNITS];
double alternative[MAX_OUTPUT_UNITS][2];
double average, target, error, extent, min, ratio, min_ratio;
double total_neg, total;
#endif

for (j=0; j<num_examples; j++)
	{
	mp = &map[j];
#ifndef PATTERN
	length = read_in_data(mp,output[MAX_DELAY]);
#else
	length = read_in_data(mp,output[MAX_DELAY],increment);
#endif
	forward_pass(length,w,output[MAX_DELAY]);
#ifdef PATTERN
	for (x=0; x<length; x++)
		{
		max = 0;
#ifdef VERBOSE
		printf("%s  ",mp->filename);
#endif
		for (out=0; out<num_output_units; out++)
			{
			unit = output_units_index[out];
			dat = output[MAX_DELAY+x][unit];
			if (dat > max)
				{
				max = dat;
				index = out;
				}
#ifdef VERBOSE
			printf("%f ",dat);
#endif
			}

#ifdef ASCII_CLASS_DATA
		printf("%d ",increment[x]);
#else
		for (out=0; out<num_output_units; out++) 
			if (mp->output_unit_sign[out] == 1.00) unit=out;
		printf("%d ",unit);
#endif
		if (num_output_units == 1)
			{
#ifdef ASCII_CLASS_DATA
			if ((max > 0.5) && (increment[x] == 0)) printf("C\n");
			else if ((max < 0.5) && (increment[x] != 0)) printf("C\n");
			else printf("X\n");
#else
			if ((max > 0.5) && ((mp->output_unit_sign[index]) == 1.000)) printf("C\n");
			else if ((max < 0.5) && (mp->output_unit_sign[index] == 0.00)) printf("C\n");
			else printf("X\n");
#endif
			}
		else
			{
			printf("%d ",index);
#if ASCII_CLASS_DATA
			if (increment[x] == index) printf("C\n");
			else printf("X\n");
#else
			if (mp->output_unit_sign[index] == 1.000) printf("C\n");
			else printf("X\n");
#endif
			}
		}
#else
	for (x=0; x<length; x++)
		{
		extent = ((double) x) / length;
		increment[x] = target_func(extent,mp->xi,
			mp->mag);
		}
/*
	Compute Error for each output unit for +/- targets
*/
	for (i=0; i<num_output_units; i++)
		{
		test_unit = output_units_index[i];
/*
	Do Plus Target
*/
		average = 0;
		for (x=0; x<length; x++)
			{
#ifdef ASYMMETRIC
			target = 0.05 + 2 * increment[x];
#else
			target = 0.500 + increment[x];
#endif
#ifdef INTERMITTENT
			if (increment[x] > 0) error = target - output[x][test_unit];
			else error = 0;
#else
			error = target - output[x][test_unit];
#endif
			average += error * error;
			}
#ifdef LMS
		if (length != 0) average/=length;
#endif
		alternative[i][PLUS] = average/2;
/*
	Do Negative Target
*/
		average = 0;
		for (x=0; x<length; x++)
			{
#ifdef ASYMMETRIC
			target = 0.05;
#else
			target = 0.500 - increment[x];
#endif
#ifdef INTERMITTENT
			if (increment[x] > 0) error = target - output[x][test_unit];
			else error = 0;
#else
			error = target - output[x][test_unit];
#endif
			average += error * error;
			}
#ifdef LMS
		if (length != 0) average/=length;
#endif
		alternative[i][MINUS] = average/2;
		}
/*
	Now compute real squared error
*/
	for (i=0; i<num_output_units; i++)
		{
		if (mp->output_unit_sign[0] == 1.00) total += alternative[i][PLUS];
		else total += alternative[i][MINUS];
		}
	total /= num_output_units;
/*
	If only one output unit, then match sign of target with
 	best score, threshold by ratio of hypotheses.
	 	compute lowest error index
	 	if matches sign then correct
	 	compute ratio
	 	if OK then correct else reject.
*/
	if (num_output_units == 1)
		{
		printf("Example %s\tHypothesis 0\t",mp->filename);
		if (mp->output_unit_sign[0] == 1.00)	/* Should be PLUS */
			{
			printf("Score %f\n",alternative[0][PLUS]);
			ratio = alternative[0][MINUS]/alternative[0][PLUS];
			if (ratio >1.00) printf("RATIO %f\tCORRECT\n",ratio);
			else printf("RATIO %f\tERROR\n",1/ratio);
			}
		else
			{
			printf("Score %f\n",alternative[0][MINUS]);
			ratio = alternative[0][PLUS]/alternative[0][MINUS];
			if (ratio >1.00) printf("RATIO %f\tCORRECT\n",ratio);
			else printf("RATIO %f\tERROR\n",1/ratio);
			}
		}
/*
	Now, determine if best hypothesis is correct
		then if expected one (CORRECT)
		else ERROR

	for each hypothesis,
		compute PLUS for that unit plus sum of MINUS
		for all other units
*/
	else
		{
		total_neg = 0;
		for (i=0; i<num_output_units; i++)
			total_neg += alternative[i][MINUS];

		for (hyp=0; hyp<num_output_units; hyp++)
			{
#ifdef LMS
			score[hyp] = (total_neg
				+ alternative[hyp][PLUS]
				- alternative[hyp][MINUS])/num_output_units;
#else
			score[hyp] = (total_neg
				+ alternative[hyp][PLUS]
				- alternative[hyp][MINUS]);
#endif

			printf("Example %s\tHypothesis %d\tScore %f\n",
				mp->filename,hyp,score[hyp]);
			}

		argmin = 0;
		min = score[0];
		for (i=1; i<num_output_units; i++)
			if (score[i] < min)
				{
				argmin = i;
				min = score[i];
				}

		min_ratio = 1000000000;
		for (i=0; i<num_output_units; i++)
			{
			ratio = score[i]/min;
			if ((i != argmin) && (ratio < min_ratio))
			  min_ratio = ratio;
			}
		printf("RATIO %f\t",min_ratio);
		if (mp->output_unit_sign[argmin] == 1.00)
			printf("CORRECT\n");
		else printf("ERROR\n");
		}
#endif
	}
}



main(argc,argv)
int argc;
char *argv[];
{
static struct vector x0;

space = initialize(argc,argv,&x0);

score_network(&x0);
}


