/*
 Author: 	Raymond L. Watrous
		Department of Computer and Information Science
		University of Pennsylvania,
		Philadelphia,PA 19104
							 December, 1988.

		Department of Computer Science
		University of Toronto,
		Toronto, Ontario M5S1A4
							 September, 1990.

 Copyright (C) 1988, Trustees of the University of Pennsylvania.
 Copyright (C) 1990, Trustees of the University of Toronto.

 Permission is granted to any individual or institution to use, copy, or
 redistribute this software so long as it is not sold for profit, provided this
 copyright notice is retained. 

			   NO WARRANTY

  Because GRADSIM is licensed free of charge, absolutely NO WARRANTY is
  provided. GRADSIM is supplied "as is" without warranty of any kind, either
  expressed or implied, including, but not limited to, the implied warranties
  of merchantability and fitness for a particular purpose. The entire risk as
  to the quality and performance of the program is with you. Should the
  program prove defective, you assume the cost of all necessary servicing,
  repair or correction.

  In no event will Raymond L. Watrous, the University of Pennsylvania,
  or the University of Toronto be liable for damages, including any lost profits,
  lost monies, or other special, incidental or consequential damages arising out of
  the use or inability to use this program.

*/

/*

	unit.c

	defines unit potential function and slope
	as well as foward pass function


	5/18/87 - modified to operate with vector data structures

	11/3/87 - reasserted delay structure for general delay

	1/6/88 - modified to be compatible with Version 1.3 network
		compiler; forward_pass from 0 to L-1, and called
		with length parameter.

	1/20/88 - modified for true LMS

*/


#include <stdio.h>
#include <math.h>
#include "net.h"
#include "vector.h"
#include "read.h"


extern int to_unit_count;
extern int from_index[],to_index[],unit_index[],delay_index[];
extern int num_weights;

#ifdef SIGPI
extern int link_type[], f2_index[], d2_index[];
#endif

extern int output_units_index[MAX_OUTPUT_UNITS];
extern int num_output_units;

double potential[MAX_NUM_UNITS];
double slope[MAX_DURATION][MAX_NUM_UNITS];
double unit_error[MAX_NUM_UNITS][MAX_DURATION + MAX_DELAY];


char net_name[MAX_NAME_LENGTH] = "NOT COMPILED";


#ifdef MULTI_UNIT_TYPE
extern int unit_type[];
#endif

#ifdef MULTI_UNIT_TYPE

double potential_function(potential,unit_index)
double potential;
int unit_index;
{
if (unit_type[unit_index] == LINEAR_UNIT) return(potential);
#ifdef EXP_OVRFLO
else if (potential < -86.0) return(0.00);
#endif
else return(1/(1 + exp(-potential)));
}


double compute_slope(x,ui)
double x;
int ui;
{
if (unit_type[ui] == LINEAR_UNIT) return(1.0000);
else return(x * (1 - x));
}

#else 

#ifdef LINEAR
double potential_function(potential)
double potential;
{
return(potential);
}

double compute_slope(x)
double x;
{
return(1.0000);
}

#else
double potential_function(potential)
double potential;
{
#ifdef EXP_OVRFLO
if (potential < -86.0) return(0.00);
#endif
return(1/(1 + exp(-potential)));
}


double compute_slope(x)
double x;
{
return(x * (1 - x));
}
#endif

#endif


int forward_pass(length,x,od)
int length;
struct vector *x;
double od[MAX_DURATION][MAX_NUM_UNITS];
/*
Accesses the output data and weight data structures
	clear potential array
	for all weights,
		accumulate weighted sum of output (delayed)
	compute outputs from potentials
Assumes that negative indices address prehistoric initial conditions
*/

{
int u,w,unit,time;
int from,to,index;
#ifdef SIGPI
int f2,ind2;
#endif
#ifdef PATTERN
int basis;
#endif

for (time=0; time<length; time++)
	{
	for (u=0; u<to_unit_count; u++)
		{
		unit = unit_index[u];
		potential[unit] = 0.00;
		}
#ifdef PATTERN
	basis = to_index[0];
	u=0;
#endif
	for (w = 0; w < num_weights; w++)
		{
		to = to_index[w];
#ifdef PATTERN
		if (to != basis)
			{
			unit = unit_index[u];
#ifdef MULTI_UNIT_TYPE
			od[time][unit] = potential_function(potential[unit],u);
#else
			od[time][unit] = potential_function(potential[unit]);
#endif
			u++;
			basis = to;
			}
#endif
		from = from_index[w];
		index = time - delay_index[w];
#ifdef SIGPI
		if (link_type[w] == SIGMA_PI)
			{
			f2 = f2_index[w];
			ind2 = time - d2_index[w];
		 	potential[to] += x->val[w] * od[index][from] * od[ind2][f2];
			}
		else potential[to] += x->val[w] * od[index][from];
#else
		potential[to] += x->val[w] * od[index][from];
#endif
		}

#ifdef PATTERN
	unit = unit_index[u];
#ifdef MULTI_UNIT_TYPE
	od[time][unit] = potential_function(potential[unit],u);
#else
	od[time][unit] = potential_function(potential[unit]);
#endif
#else
	for (u=0; u<to_unit_count; u++)
		{
		unit = unit_index[u];
#ifdef MULTI_UNIT_TYPE
		od[time][unit] = potential_function(potential[unit],u);
#else
		od[time][unit] = potential_function(potential[unit]);
#endif
		}
#endif
	}
}





int clear_unit_error()
{
int i,j;

for (i=0; i<MAX_NUM_UNITS; i++)
	for (j=0; j<MAX_DURATION+MAX_DELAY; j++)
		unit_error[i][j] = 0.00;
}




#ifdef SIGPI
back_propagate(weight,length,od)
double od[MAX_DURATION][MAX_NUM_UNITS];
#else
back_propagate(weight,length)
#endif
struct vector *weight;
int length;
/*
Error is back-propagated along incoming links, now by iteration.
*/
{
int u,unit,time;
int to,from,w,index;
#ifdef SIGPI
int f2,ind2;
#endif
#ifdef PATTERN
int basis;
#endif

for (time=length-1; time>=0; time--)
	{
#ifndef PATTERN
	for (u=0; u < to_unit_count; u++)
		{
		unit = unit_index[u];
		unit_error[unit][time+MAX_DELAY] *= slope[time][unit];
		}
	for (w=0; w<num_weights; w++)
		{
#else
	u = to_unit_count - 1;
	basis = -100;

	for (w=num_weights-1; w>=0; w--)
		{
		to = to_index[w];
		if (to != basis)
			{
			unit = unit_index[u--];
			unit_error[unit][time+MAX_DELAY] *= slope[time][unit];
			basis = to;
			}
#endif
		from = from_index[w];
		index = time - delay_index[w];
#ifdef SIGPI
		if (link_type[w] == SIGMA_PI)
			{
			f2 = f2_index[w];
			ind2 = time - d2_index[w];
			to = to_index[w];

			if (from >= INPUT_UNITS)
				{
				unit_error[from][index+MAX_DELAY] +=
					unit_error[to][time+MAX_DELAY] * 
					  od[ind2][f2] *
					    weight->val[w];
				}
			if (f2 >= INPUT_UNITS)
				{
				unit_error[f2][ind2+MAX_DELAY] +=
					unit_error[to][time+MAX_DELAY] * 
					  od[index][from] *
					    weight->val[w];
				}
			}
		else if (from >= INPUT_UNITS)
			{
			to = to_index[w];
			unit_error[from][index+MAX_DELAY] +=
				unit_error[to][time+MAX_DELAY] * 
					weight->val[w];
			}
#else
		if (from >= INPUT_UNITS)
			{
#ifndef PATTERN
			to = to_index[w];
#endif
			unit_error[from][index+MAX_DELAY] +=
				unit_error[to][time+MAX_DELAY] * 
					weight->val[w];
			}
#endif
		}
	}
}



accumulate_error_vect(vect,len,od)
struct vector *vect;
int len;
double od[MAX_DURATION][MAX_NUM_UNITS];
/*
Accumulate_error performs the weight change accumulation, for all time
	CHANGE ORDER OF INDEXING!!!!
*/
{
int w,ref,to,from,time;
#ifdef SIGPI
int ref2, f2;
#endif

for (time=0; time<len; time++)
	for (w=0; w<num_weights; w++)
		{
		ref = time - delay_index[w];
		to = to_index[w];
		from = from_index[w];
#ifdef SIGPI
		if (link_type[w] == SIGMA_PI)
			{
			ref2 = time - d2_index[w];
			f2 = f2_index[w];
			vect->val[w] -= unit_error[to][time+MAX_DELAY]
				 * od[ref][from]
				 * od[ref2][f2];
			}
		else vect->val[w] -= unit_error[to][time+MAX_DELAY]
				 * od[ref][from];
#else
		vect->val[w] -= unit_error[to][time+MAX_DELAY]
				 * od[ref][from];
#endif
		}
}





int add_avg_vect(dst,src,count)
double *dst, *src;
int count;
/*
	WARNING - GUARD AGAINST DIVIDE BY ZERO
*/
{
register double scale;
register struct vector *dp,*sp;
register int i;

if (count == 0) return(0);
scale = count;
for (i=0; i<num_weights; i++)
	*dst++ += *src++/scale;
}



cg3 (od, grad, weight, ds, length, inc)
double od[MAX_DURATION][MAX_NUM_UNITS];
struct vector *grad, *weight;
struct map_entry *ds;
int length;
#ifndef PATTERN
double inc[];
#else
long int inc[];
#endif
/*
Assumes that forward pass has been appropriately called
*/
{
register double target;
int out,unit,u,time;
struct vector partial;
double epsilon;

for (time=0; time<length; time++)
	{
	for (u=0; u<to_unit_count; u++)
		{
		unit = unit_index[u];
#ifdef MULTI_UNIT_TYPE
		slope[time][unit] = compute_slope(od[time][unit],u);
#else
		slope[time][unit] = compute_slope(od[time][unit]);
#endif
		}
	}

clear_unit_error();

for (time=0; time<length; time++)
	{
#ifdef INTERMITTENT
	if (inc[time] >= 0.000) {
#endif
	for (out=0; out<num_output_units; out++)
		{
		unit = output_units_index[out];

#ifndef PATTERN
#ifdef ASYMMETRIC
		if (ds->output_unit_sign[out] == 1.00)
			target = 0.05 + 2 * inc[time];
		else target = 0.05;
#else
		if (ds->output_unit_sign[out] == 1.00)
			target = 0.500 + inc[time];
		else target = 0.500 - inc[time];
#endif
#else
#ifdef EXPLICIT
		target = ds->output_unit_sign[out];
#else
#ifdef ASCII_CLASS_DATA
		if (inc[time] == out) target = 0.9000000;
		else target = 0.100000;
#else
		if (ds->output_unit_sign[out] == 1.00) target = 0.9000000;
		else target = 0.100000;
#endif
#endif
#endif

#ifdef RADIAL
		unit_error[unit][MAX_DELAY+time] = (target - od[time][unit])/ds->xi;
#else
		unit_error[unit][MAX_DELAY+time] = (target - od[time][unit]);
#endif
		}
#ifdef INTERMITTENT
}
#endif
	}
/*
NOW, BACK PROPAGATE RIGHT-TO-LEFT
*/
#ifdef SIGPI
back_propagate(weight,length,od);
#else
back_propagate(weight,length);
#endif
#ifdef LMS
clear_vect(&partial);
accumulate_error_vect(&partial,length,od);
add_avg_vect(grad->val,&partial.val[0],length * num_output_units);
#else
accumulate_error_vect(grad,length,od);
#endif
}

