/*
 Author: 	Raymond L. Watrous
		Department of Computer and Information Science
		University of Pennsylvania,
		Philadelphia,PA 19104
							 December, 1988.

		Department of Computer Science
		University of Toronto,
		Toronto, Ontario M5S1A4
							 September, 1990.

 Copyright (C) 1988, Trustees of the University of Pennsylvania.
 Copyright (C) 1990, Trustees of the University of Toronto.

 Permission is granted to any individual or institution to use, copy, or
 redistribute this software so long as it is not sold for profit, provided this
 copyright notice is retained. 

			   NO WARRANTY

  Because GRADSIM is licensed free of charge, absolutely NO WARRANTY is
  provided. GRADSIM is supplied "as is" without warranty of any kind, either
  expressed or implied, including, but not limited to, the implied warranties
  of merchantability and fitness for a particular purpose. The entire risk as
  to the quality and performance of the program is with you. Should the
  program prove defective, you assume the cost of all necessary servicing,
  repair or correction.

  In no event will Raymond L. Watrous, the University of Pennsylvania,
  or the University of Toronto be liable for damages, including any lost profits,
  lost monies, or other special, incidental or consequential damages arising out of
  the use or inability to use this program.

*/

/*
	weight.c -

		Reads in network link-weight descriptor


	r. watrous		september 25, 1986

	modified 10/30/87
		allow for gaps in unit space

	modified 1/31/89
		for sigma_pi links
*/

#include <stdio.h>
#include "net.h"

#undef DEBUG

extern int threshold_unit;

extern int to_unit_count;


store_weights(name,wset)
char *name;
struct link *wset[];
{
struct link *lp;
FILE *fp;
int unit;

if ((fp = fopen(name,"w")) == NULL)
	{
	fprintf(stderr,"UNABLE TO OPEN LINK FILE FOR WRITING\n");
	exit(1);
	}
fprintf(fp,"Threshold Unit %d\n",threshold_unit);
for (unit=0; unit<to_unit_count; unit++)
	{
	lp = wset[unit];
	while (lp != NULL)
		{
#ifdef SIGPI
		if (lp->type == SIGMA_PI)
		fprintf(fp,"%d\t%d\t%d\t%d\t%d\t%.10f\n",lp->from_unit,lp->from2_unit,lp->to_unit,
				lp->delay,lp->delay2,lp->weight);
		else fprintf(fp,"%d\t\t%d\t%d\t\t%.10f\n",lp->from_unit,lp->to_unit,
				lp->delay,lp->weight);
#else
		fprintf(fp,"%d\t%d\t%d\t%.10f\n",lp->from_unit,lp->to_unit,
			lp->delay,lp->weight);
#endif
		lp = lp->next;
		}	
      }
fclose(fp);
}



int compare(lp1,lp2)
struct link *lp1,*lp2;
{
int diff;

diff = 0;

if (lp1->to_unit != lp2->to_unit) diff = 1;
if (lp1->from_unit != lp2->from_unit) diff = 1;
if (lp1->delay != lp2->delay) diff = 1;
#ifdef SIGPI
if (lp1->type != lp2->type) diff = 1;
else if (lp1->type == SIGMA_PI)
	{
	if (lp1->from2_unit != lp2->from2_unit) diff = 1;
	if (lp1->delay2 != lp2->delay2) diff = 1;
	}
#endif
return(diff);
}


cons(lp,lplist)
struct link *lp,*lplist;
{
if (lplist == NULL)
	{
	fprintf(stderr,"***NULL LINK LIST***\n");
	exit(2);
	}
else if (!compare(lp,lplist))
	{
	fprintf(stderr,"IDENTICAL LINKs READ\t");
	fprintf(stderr,"%d\t%d\t%d\n",lp->from_unit,lp->to_unit,lp->delay);
	exit(1);
	}
while (lplist->next != NULL)
	{
	lplist = lplist->next;
	if (!compare(lp,lplist))
		{
		fprintf(stderr,"IDENTICAL LINKs READ\t");
		fprintf(stderr,"%d\t%d\t%d\n",lp->from_unit,lp->to_unit,
				lp->delay);
		exit(1);
		}
	}
lplist->next = lp;
}


#ifdef SIGPI

int read_link(fp,from,to,delay,f2,d2,weight)
FILE *fp;
int *from,*to,*delay,*f2,*d2;
float *weight;
/*
Read link descriptor to handle problem with -0.35, for example
*/
{
int if1,if2,if3,if4,if5;
float ff1;
int x,type;
char s,c[10];

if ((x = fscanf(fp,"%d %d %d",&if1,&if2,&if3)) == EOF) return(x);

/*
If negative sign, then must be WEIGHT for simple link
*/
fscanf(fp,"%1s",c);
s = c[0];
if (s == '-')
	{
	x = fscanf(fp,"%f\n",&ff1);
	*weight = -ff1;
	*from = if1;
	*to = if2;
	*delay = if3;
	type = SIMPLE;
	}	
else
	{
	ungetc(s,fp);
	x = fscanf(fp,"%d %d %f\n",&if4,&if5,&ff1);
	if (x == 3)
		{
		*from = if1;
		*f2 = if2;
		*to = if3;
		*delay = if4;
		*d2 = if5;
		*weight = ff1;
		type = SIGMA_PI;
		}
	else
		{
/*
Hebb now works differently; it will read if4 as integer, balk at decimal if5,
but PROCEED TO READ FLOAT INTO ff1!! returning 2 for x
*/
		if (x==1) fscanf(fp,"%f\n",&ff1);
		else if (x!=2) fprintf(stderr,"FUNKY READ %d\n",x);
#ifdef DEBUG
		fprintf(stderr,"read %d fields: %d %d %f\n",x,if4,if5,ff1);
#endif
		*from = if1;
		*to = if2;
		*delay = if3;
		*weight = (float) if4 + ff1;
		type = SIMPLE;
		}
	}
if (type == SIMPLE) return(4);
else if (type == SIGMA_PI) return(6);
else return(EOF);
}
#endif



read_weights(name,wset)
char *name;
struct link *wset[];
/*
	checks that highest to_unit is less than MAX_NUM_UNITS
*/
{
FILE *fp;
int from,to,delay;
#ifdef SIGPI
int f2,d2,type;
#endif
float weight;
struct link *nlp;
int x;
int found,index,to_cnt;


if ((fp = fopen(name,"r")) == NULL)
	{
	fprintf(stderr,"UNABLE TO OPEN LINK FILE FOR READING\n");
	exit(1);
	}
to_cnt = 0;
if ((x = fscanf(fp,"Threshold Unit %d",&threshold_unit)) != 1)
	{
	fprintf(stderr,"***COULD NOT READ THRESHOLD UNIT***\n");
	fprintf(stderr,"***IN FILE %s***\n",name);
	exit(1);
	}
#ifdef SIGPI
while ((x = read_link(fp,&from,&to,&delay,&f2,&d2,&weight)) != EOF)
	{
	if (x == 4) type = SIMPLE;
	if (x == 6) type = SIGMA_PI;
#else
while ((x = fscanf(fp,"%d %d %d %f\n",&from,&to,&delay,&weight)) == 4)
	{
#endif
#ifdef DEBUG

#ifdef SIGPI
		if (type == SIGMA_PI)
		fprintf(stderr,"%d\t%d\t%d\t%d\t%d\t%.10f\n",from,f2,to,delay,d2,weight);
		else fprintf(stderr,"%d\t\t%d\t%d\t\t%.10f\n",from,to,delay,weight);
#else
		fprintf(fp,"%d\t%d\t%d\t%.10f\n",from,to,delay,weight);
#endif

#endif

	if (to >= MAX_NUM_UNITS)
		{
		fprintf(stderr,"***TO UNIT INDEX EXCEEDS MAXIMUM***\n");
		fprintf(stderr,"File %s - unit %d\n",name,to);
		exit(1);
		}
	if (from >= MAX_NUM_UNITS)
		{
		fprintf(stderr,"***FROM UNIT INDEX EXCEEDS MAXIMUM***\n");
		fprintf(stderr,"File %s - unit %d\n",name,from);
		exit(1);
		}
	if (to == threshold_unit)
		{
		fprintf(stderr,"***THRESHOLD UNIT USED AS TO INPUT***\n");
		fprintf(stderr,"File %s - unit %d\n",name,to);
		exit(1);
		}
	if (delay > MAX_DELAY)
		{
		fprintf(stderr,"Delay %d exceeds MAX_DELAY - leaving\n",
			delay);
		exit(1);
		}
#ifdef SIGPI
	if ((type == SIGMA_PI) && (d2 > MAX_DELAY))
		{
		fprintf(stderr,"Delay %d exceeds MAX_DELAY - leaving\n",
			d2);
		exit(1);
		}
#endif

	nlp = (struct link *) malloc(sizeof(struct link));
	nlp->from_unit = from;
	nlp->to_unit = to;
	nlp->delay = delay;
	nlp->weight = weight;
	nlp->next = NULL;
#ifdef SIGPI
	nlp->from2_unit = f2;
	nlp->delay2 = d2;
	nlp->type = type;
#endif

/*
	if (wset[to - INPUT_UNITS] == NULL) wset[to - INPUT_UNITS] = nlp;
	else cons(nlp,wset[to - INPUT_UNITS]);
*/

/*	Add link to list	
		scan array until to_unit matched or end
		if match - add
		if fail	- add new and increment count
*/

	found = 0; index = 0;
	while ((found==0) && (wset[index] != NULL))
		{
		if (wset[index]->to_unit == to)
			{
			found = 1;
			cons(nlp,wset[index]);
			}
		index++;
		}
	if (found == 0)
		{
		wset[index] = nlp;
		to_cnt++;
		}
	}
if (to_cnt == 0)
	{
	fprintf(stderr,"***TOP COUNT = 0 - CHECK LINK FILE***\n");
	fprintf(stderr,"File %s - unit %d\n",name,to_cnt);
	exit(1);
	}
fclose(fp);
return(to_cnt);
}

#ifdef MULTI_UNIT_TYPE

/*
        read_unit_type.c
                reads in unit_type descriptor

*/

#include "vector.h"
#include "read.h"

extern int unit_type[];

int read_unit_type_descriptor(name,link_set,count)
char *name;
struct link *link_set[];
int count;
{
int nrd, units;
int unit, type;
FILE *fp;
struct vector *vp;
int unit_ndx;
struct link *lp;

if ((fp = fopen(name,"r")) == NULL)
	{
	printf("CANNOT OPEN UNIT TYPE DESCRIPTOR");
	exit(1);
	}

units = 0;

while (((nrd = fscanf(fp,"%d %d",&unit,&type)) != EOF) && (units<MAX_NUM_UNITS))
	{
	for (unit_ndx=0; unit_ndx <count; unit_ndx++)
		{
		lp = link_set[unit_ndx];
		if (unit == lp->to_unit) unit_type[unit_ndx] = type;
		}
	units++;
	}
fclose(fp);
return(units);
}

#endif

